/**
 *@file LOST.cc
 *
 *All code for detecting uninterrupted, tightly-connected community
 * evolutions in temporal social networks. Procedure is detected
 * in Mark Goldberg, Malik Magdon-Ismail, and James
 * Thompson. 2012. Identifying long lived social communities using
 * structural properties. In Proceedings of the 2012 International 
 * Conference on Advances in Social Networks Analysis and Mining
 * (ASONAM 2012). IEEE, 647-653.
 *
 * TODO:
 * 		Test
 * 		Input/Output/Example here
 * 	    Licensing
 * 
 *@author James Thompson
 *Copyright James Thompson 2015
 *This program is distributed under the terms of the GNU General Public License
 *
 * This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */
 
#include "../Libraries/Files/IOX.h"
#include "../Libraries/Random/Wrappers.h"
#include "../Libraries/Params/Parameters.h"
#include <set>
#include <map>
#include <vector>
#include <iostream>
#include <fstream>
#include <tuple>

typedef set < string > Community;
typedef map < string, vector < int > > Membership; //Vertex to indices of communities it's a member of

/**
 *@fn void read_file ( const string& filename, vector < Community >& structure, Membership& membership_structure )
 *
 *Takes a file (filename) and reads in the community structure 
 *   described within. 
 *
 *@param filename Name of file holding a community structure. Each
 *                  line should be a single community, members 
 *                  space delimited. 
 *@param structure List of the groups of members to be considered
 *                   communities read from given file.
 *@param membership_structure Map of vertices to id of communities 
 *                              that vertex is a member of.
 */
void read_file ( const string& filename, vector < Community >& structure, Membership& membership_structure, set < string >& unk );

/**
 *@fn double Jaccard ( const Community& A, const Community& B )
 * 
 *Computes the Jaccard Index between memberships of two communities
 *   For use in calculating single valid evolutionary steps.
 *
 *@param A,B Communities to compare
 *@return Number of shared members divided by the total number of unique 
 *          members in the union of both communities.
 */
double Jaccard ( const Community& A, const Community& B, const double& num_unk = 0 );

/**
 * @fn double Jaccard_Naive_Unknown ( const Community& A, const Community& B, const double& num_unk = 0 )
 * 
 * Calculates the Jaccard index between communities with possibly unknown
 * members by ignoring them - two unknown members are assumed to not
 * match.
 * 
 * @param A,B Communities to be compared
 * @param num_unk Total number of unknown vertices in the network
 */
double Jaccard_Naive_Unknown ( const Community& A, const Community& B, const double& num_unk = 0 );

/**
 * @fn double Jaccard_Naive_Probabilistic_Unknown ( const Community& A, const Community& B, const double& num_unk )
 * 
 * Calculates the Jaccard Index between communities with possible unknowns.
 * Using the total number
 * of unknowns, the expected number of matches of unknowns between 
 * communities if each member is equally likely to match any other is 
 * added to the number of matches of known vertices.
 * 
 * @param A,B Communities to be compared
 * @param num_unk Total number of unknown vertices in the network
 */
double Jaccard_Naive_Probabilistic_Unknown ( const Community& A, const Community& B, const double& num_unk );

/**
 * @fn double Jaccard_Naive_Static_Probabilistic_Unknown ( const Community& A, const Community& B, const double& num_unk )
 * 
 * A static probability of two unknowns matching is used to calculate
 * the Jaccard Index between two communities with possible unknowns.
 * 
 * @param A,B Communities to be compared
 * @param num_unk Total number of unknown vertices in the network
 */
double Jaccard_Naive_Static_Probabilistic_Unknown ( const Community& A, const Community& B, const double& num_unk );

/**
 *@fn double MinNorm ( const Community& A, const Community& B)
 *
 *Computes a similarity measure similar to Jaccard Index between two
 *   communities ( minimum community size is substituted for denominator )
 *
 *@param A,B Communities to compare
 *@return Number of shared vertices divided by the size of the smaller community
 */
double MinNorm ( const Community& A, const Community& B, const double& num_unk);

/**
 *@fn void Print ( const Membership& M )
 *@fn void Print ( const Community& C )
 *
 *Prints the string representation of a community/membership map to
 *   standard out. Communities are space-delimited list of members.
 *   Membership map has one vertex per line, followed by list of community
 *   IDs that it's a member of.
 *
 *@param M/C Object to print to standard out
 */
void Print ( const Membership& M );
void Print ( const Community& C );

/**
 *@fn void calculate_valid_steps ( const vector < vector < Community > >& C, const vector < Membership >& M, vector < vector < int > >& steps, double threshold = 0.15, double (*compare_fnc) ( const Community& A, const Community& B ) = &Jaccard )
 *
 *Holds main functionality of evolution detection. First calculates valid 1-step
 *  evolutions using compare_fnc and C/M parameters. Then goes back and constructs
 *  full evolutions using the valid steps.
 *
 *@param C List of community structures representing the static networks
 *@param M List of membership maps corresponding to the community structures
 *@param steps Place to hold information about valid evolution transitions
 *              steps[1][2] holds the id of the community that the second
 *              community at time window 1 would evolve into in time window 2
 *@param threshold Minimum valid value for the comparison function between two communities 
 *@param compare_fnc Function to use for comparing communities in consecutive time
 *                   windows
 */
void calculate_valid_steps ( const vector < vector < Community > >& C, const vector < Membership >& M, vector < vector < int > >& steps, double threshold = 0.15, const double& num_unk = 0, double (*compare_fnc) ( const Community& A, const Community& B, const double& num_unk ) = &Jaccard_Naive_Static_Probabilistic_Unknown );

/**
 *@fn tuple<int, int> int_tuple ( int a, int b )
 *
 *Simply converts two integer values to a tuple.
 *
 *@param a,b Integers to put into a tuple structure
 *@return A tuple where the first entry is less than 
 *            the second.
 */
tuple<int, int> int_tuple ( int a, int b ){
  if ( a < b ){
    return tuple<int, int> ( a, b );
  } else {
    return tuple<int, int> ( b, a );
  }
}

//Variable for static probability ( optionally used )
//Terrible design - oops.
static double static_unknown_prob;

/**
 *@fn int main ( int argc, char** argv )
 *
 *Insertion point for the evolution detection algorithm.
 *
 *@param argc Number of command line arguments
 *@param argv Command line arguments
 */
int main ( int argc, char** argv ){
  //Initializes structures and reads command line arguments
  Parameters P; P.Read ( argc, argv );
  ifstream fin;
  vector < vector < Community > > dyn_structure;
  vector < Membership > mem_structure;
  openFileHarsh ( &fin, P.get < string > ( "i", "communities.dat" ) );
  vector < string > filenames;

  Community A, B;   
  set < string > unknowns; 

  //Reads in the static community structures from a list of 
  //   files found in communities.dat
  while ( fline_tr ( &fin, &filenames, " " ) ){
    for ( int i = 0; i < filenames.size(); i++ ){
      vector < Community > empty;
      map < string, vector < int > > empty_member;

      read_file ( filenames[i], empty, empty_member, unknowns );

      dyn_structure.push_back ( empty );
      mem_structure.push_back ( empty_member );
    }
  }

  fin.close(); 

  //Constructs the evolutions
  vector < vector < int > > steps;
  
  double cThresh = P.get < double > ( "thr", 0.15 );
  switch ( P.get < char > ( "comp", 'j' ) ){
	case 'n':
		calculate_valid_steps ( dyn_structure, mem_structure, steps, cThresh, unknowns.size(), &Jaccard_Naive_Unknown );
		break;
	case 'p':
		calculate_valid_steps ( dyn_structure, mem_structure, steps, cThresh, unknowns.size(), &Jaccard_Naive_Probabilistic_Unknown );
		break;
	case 's':
		static_unknown_prob = P.get < double > ( "stpr", 0.3 );
		calculate_valid_steps ( dyn_structure, mem_structure, steps, cThresh, unknowns.size(), &Jaccard_Naive_Static_Probabilistic_Unknown );
		break;
	case 'm':
		calculate_valid_steps ( dyn_structure, mem_structure, steps, cThresh, unknowns.size(), &MinNorm );
		break;
	default:
		calculate_valid_steps ( dyn_structure, mem_structure, steps, cThresh, unknowns.size(), &Jaccard );
		break;
  }
}

void Print ( const Community& C ){
  //Goes through members and prints out each one 
  //   with spaces in between each
  Community::const_iterator it_c;
  for ( it_c = C.begin(); it_c != C.end(); it_c++ ){
    cout << *it_c << " ";
  }
  cout << endl;
}

void Print ( const Membership& M ){
  //Prints out each vertex followed by a list of communities
  // it is a part of.
  Membership::const_iterator it_m;
  for ( it_m = M.begin(); it_m != M.end(); it_m++ ){
    cout << it_m->first << " :";
    for ( int i = 0; i < it_m->second.size(); i++ ){
      cout << " " << it_m->second[i];
    }
    cout << endl;
  }
}

/**
 *@fn void Print ( const int i, const int j, const vector < vector < Community > >& structure, const vector < vector < int > >& steps, set < tuple < int, int > >& visited )
 *
 *@param i Time window to start at
 *@param j Community to start at
 *@param structure Community structures of static networks
 *@param steps Transition data representing valid evolutionary paths
 *@param visited Set of communities whose following evolutions have already 
 *               been printed ( possible as a sub-evolution of another community )
 */
void Print ( const int i, const int j, const vector < vector < Community > >& structure, const vector < vector < int > >& steps, set < tuple < int, int > >& visited ) {
  int ci = i, cj = j;
 
  //Prints the first community 
  Print ( structure[ci][cj] );

  //Follows the links in steps to print following communities in evolution
  while ( ( steps[ci][cj] >= 0 ) ){
    //cout << (ci + 1) << " " << steps[ci][cj] << " " << ci << " " << cj << " " << endl;
    Print ( structure[ci + 1][steps[ci][cj]] );
    cj = steps[ci][cj];
    ci = ci + 1;
    
    //Makes sure sub-evolutions are not reprinted
    visited.insert ( tuple < int , int > ( ci, cj ) );
  }

  cout << "------------" << endl;
}
 
void read_file ( const string& filename, vector < Community >& structure, Membership& membership_structure, set < string >& unk ){
  ifstream fin;
  openFileHarsh( &fin, filename );
  
  Community next_com;
  Community::iterator it_c;
  Membership::iterator it_m;
  
  while ( fline_tr ( &fin, &next_com, " " ) ){    
    for ( it_c = next_com.begin(); it_c != next_com.end(); it_c++ ){
	  //Tracks the number of unknowns
      if ( (*it_c)[0]=='U' ) {
	//cerr << "inknoen" << endl;
	unk.insert ( *it_c );
      }
		
	//Adds community to member map of all members
      if ( ( it_m = membership_structure.find ( *it_c ) ) != membership_structure.end() ){
	it_m->second.push_back( structure.size() );
      } else {
	vector < int > tmp; tmp.push_back(structure.size());
	membership_structure.insert(pair <string, vector < int > > ( *it_c, tmp ) );
      }
    }
    
    //Adds community to structure
    structure.push_back ( next_com );
    next_com.clear();
  }

  fin.close();
}

double Jaccard ( const Community& A, const Community& B, const double& num_unk ){
  double same = 0.0;
  double total = 0.0;

  //Calculates similar members in each community
  Community::const_iterator it_a, it_b;
  for ( it_a = A.begin(); it_a != A.end(); it_a++ ){
    ++total;
    if ( B.find ( *it_a ) != B.end() ) ++same;
  }

  //Calculates the unique members in B, adding them 	
  // to the count in A, resulting in total number of unique members
  for ( it_b = B.begin(); it_b != B.end(); it_b++ ){
    if ( A.find ( *it_b ) == A.end() ) ++total;
  }

  return same / total;
}

double Jaccard_Naive_Unknown ( const Community& A, const Community& B, const double& num_unk ){
  //static int call_count = 0;
  double same = 0.0;
  double total = 0.0;
  double unknown = 0.0;
  
  Community::const_iterator it_a, it_b;
  
  //Tabulates the number of shared, unknown and unique members of A
  for ( it_a = A.begin(); it_a != A.end(); it_a++ ){
    if ( ( (*it_a)[0] == 'U' ) || ( (*it_a)[0] == 'u' ) ){
      ++total;
      ++unknown;
    } else {
      if ( B.find ( *it_a ) != B.end() ){
	++same;
      }
      
      ++total;
    }
  }
	
  //Tabulates shared,unknown, and unique members of B
  for ( it_b = B.begin(); it_b != B.end(); it_b++ ){
    if ( ( (*it_b)[0] == 'U' ) || ( (*it_b)[0] == 'u' ) ){
      ++total;
      ++unknown;
    } else {
      if ( A.find ( *it_b ) == A.end() ){
	++total;
      }
    }
  }
  
  /**cout << "Unknown: " << unknown << endl;
  cout << "Same: " << same << endl;
  cout << "Total " << total << endl;
  cout << "Normal Jaccard : " << Jaccard ( A, B, num_unk ) << endl;
  cout << "Naive Unknown Jaccard : " << same / total << endl;
  cout << "=======================" << endl;
  if ( ++call_count == 10 )
  exit ( 10 );**/
  //Assumes all unknowns are different, and returns classical Jaccard 
  return same / total;
}

double Jaccard_Naive_Probabilistic_Unknown ( const Community& A, const Community& B, const double& num_unk ){
  double same = 0.0;
  double total = 0.0;
  double unknown_a = 0.0;
  double unknown_b = 0.0;
  
  Community::const_iterator it_a, it_b;
  
  //Tabulates the number of shared, unknown and unique members of A
  for ( it_a = A.begin(); it_a != A.end(); it_a++ ){
    if ( ( (*it_a)[0] == 'U' ) || ( (*it_a)[0] == 'u' ) ){
      ++total;
      ++unknown_a;
    } else {
      if ( B.find ( *it_a ) != B.end() ){
	++same;
      }
      
      ++total;
    }
  }

  //Tabulates the number of shared, unknown and unique members of B
  for ( it_b = B.begin(); it_b != B.end(); it_b++ ){
    if ( ( (*it_b)[0] == 'U' ) || ( (*it_b)[0] == 'u' ) ){
      ++total;
      ++unknown_b;
    } else {
      if ( A.find ( *it_b ) == A.end() ){
	++total;
      }
    }
  }

  //Calculates the expected number of matches between unknowns
  double unk_match = ( (unknown_a * unknown_b) / num_unk );
  //Print ( A );
  //Print ( B );
  //cout << same << " " << total << " " << unk_match << " " << num_unk << " " << ( (same + unk_match) / ( total - unk_match ) ) << endl;
  return ( (same + unk_match) / ( total - unk_match ) );
}

double Jaccard_Naive_Static_Probabilistic_Unknown ( const Community& A, const Community& B, const double& num_unk ){
  double same = 0.0;
  double total = 0.0;
  double unknown_a = 0.0;
  double unknown_b = 0.0;
  
  Community::const_iterator it_a, it_b;
  
  //Tabulates the number of shared, unknown and unique members of A
  for ( it_a = A.begin(); it_a != A.end(); it_a++ ){
    if ( ( (*it_a)[0] == 'U' ) || ( (*it_a)[0] == 'u' ) ){
      ++total;
      ++unknown_a;
    } else {
      if ( B.find ( *it_a ) != B.end() ){
	++same;
      }
      
      ++total;
    }
  }
  //Tabulates the number of shared, unknown and unique members of B
  for ( it_b = B.begin(); it_b != B.end(); it_b++ ){
    if ( ( (*it_b)[0] == 'U' ) || ( (*it_b)[0] == 'u' ) ){
      ++total;
      ++unknown_b;
    } else {
      if ( A.find ( *it_b ) == A.end() ){
	++total;
      }
    }
  }

  //Go through and match unknowns using a static probability
  int random_matches = 0;
  for ( int i = 0; i < unknown_a; i++ ){
    for ( int j = 0; j < unknown_b; j++ ){
      if ( random_double() < static_unknown_prob ){ 
		++random_matches;
		--unknown_b;
		break;
      }
    }
  }

  //Calculates the Jaccard Index
  return ( same + random_matches ) / ( total - random_matches );
}

double MinNorm ( const Community& A, const Community& B, const double& num_unk ){
  double same = 0.0;

  //Calculates similar members in each community
  Community::const_iterator it_a, it_b;
  for ( it_a = A.begin(); it_a != A.end(); it_a++ ){
    if ( B.find ( *it_a ) != B.end() ) ++same;
  }

  double smaller_size = (A.size() < B.size()) ? A.size() : B.size(); 

  return same / smaller_size;
}

/**
 * @fn int path_length ( const int start_window, const int start_com, const vector < vector < int > >& steps )
 * 
 * Calculates the length of the valid evolution that starts from the 
 * 	start_com community in the start_window time window.
 * 
 * @param start_window Time window of community to start from
 * @param start_com Community ID to start from
 * @param steps Transition data for evolutions
 */
int path_length ( const int start_window, const int start_com, const vector < vector < int > >& steps ){
  int i = start_window, j = start_com;
  int len = 0;
  while ( (i < steps.size() ) && ( j < steps[i].size() ) && (i > 0 ) && (j>0) && ( steps[i][j] >= 0 ) ) {
    j = steps[i][j];
    i++;
    ++len;
  }

  return len;
}

void calculate_valid_steps ( const vector < vector < Community > >& C, const vector < Membership >& M, vector < vector < int > >& steps, double threshold, const double& num_unk, double (*compare_fnc) ( const Community& A, const Community& B, const double& num_unk ) ) {
  //cerr << "UNKNOWNS!!! + " << num_unk << endl;
  steps.clear();
  for ( int i = 0; i < C.size(); i++ ){
    vector < int > init ( C[i].size(), -1 );
    steps.push_back( init );
  }
  
  set < tuple < int, int > > matches;
  Membership::const_iterator it_m, it_n;

  for ( int i = C.size() - 2; i >= 0; i-- ){
	//Retreives all pairs of communities, one from time window i, one from i+1
	//    which have at least one member similar
	matches.clear();
	for ( it_m = M[i].begin(); it_m != M[i].end(); it_m++ ){
	  if ( ( it_n = M[i + 1].find ( it_m->first ) ) != M[i+1].end() ){
		for ( int k = 0; k < it_m->second.size(); k++ ){
		  for ( int j = 0; j < it_n->second.size(); j++ ){
			matches.insert ( tuple <int, int> ( it_m->second[k], it_n->second[j] ) );
		  }
		}
	  }
	}

	//Goes through all possible pairs of communities ( with at least one
	//     similar member ) and calculates the similarity measure. Each
	//     community is only connect to the community in the following window
	//     with the highest similarity ( or longest evolution if tied ( or random 
	//     if still tied ) )
    set < tuple < int, int > >::iterator it_t;
    it_t = matches.begin();
    while ( it_t != matches.end() ) {
      int current_base = get < 0 > ( *it_t );
      int best_target = -1;      
      double best_value = -1.0, next_value;
      while ( ( it_t != matches.end() ) && ( get < 0 > ( *it_t ) == current_base ) ){
		if ( ( ( next_value = compare_fnc ( C[i][get<0>(*it_t)], C[i+1][get<1>(*it_t)], num_unk ) ) >= best_value ) && ( next_value > threshold ) ) {
		  if ( next_value == best_value ){
			if ( path_length ( i+1, best_target, steps ) < path_length ( i+1, get < 1 > (*it_t), steps ) ) {
			  best_value = next_value;
			  best_target = get<1>(*it_t);	      
			}
		  } else {  
			best_value = next_value;
			best_target = get<1>(*it_t);
		  }
	}

	//cout << next_value << " " << get < 0 > (*it_t ) << " " << get < 1 > (*it_t) << endl;

	++it_t;
      }
	
		//Tracks the valid evolution steps
      if ( best_value > -1 )
		steps[i][current_base] = best_target;
	  }
  }

  //Prints out full evolutions using single valid steps
  set < tuple < int, int > > visited;

  for ( int i = 0; i < steps.size(); i++ ){
    cout << "Timestep " << i << endl;
    for ( int j = 0; j < steps[i].size(); j++ ){
      if ( visited.find(tuple < int, int > (i, j) ) == visited.end() ) {
	//Print out the evolution for i,j
	Print ( i, j, C, steps, visited); 
      }
    }
  }
}
