/**
 *@file extend_evolutions.cc
 * 
 * Program to extend short, uninterrupted evolutions to longer, possibly
 * intermittant ones.
 * 
 * The uninterrupted evolutions should be in a file where each evolution
 * is separated by a line of dashes (-). The evolutions starting at time
 * step 0 should precede those starting at time step 1, which precedes
 * those starting from time step 2, etc. To denote the change in the 
 * starting time window, the file should have a line that says
 * 'Timestep #', where # is replaced with the approriate number. See
 * examples folder for example.
 * 
 * @author James Thompson
 * Copyright James Thompson 2015
 *
 *This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include "../Libraries/Params/Parameters.h"
#include "../Libraries/Files/IOX.h"
#include <map>
#include <cmath>

#include <tr1/memory>

using namespace std;

typedef vector < string > Community;
typedef vector < Community > Evolution;

/**
 * @fn map < string, int > get_counts ( Evolution e )
 * 
 * The M-value (evolution comparison) used depends on participation
 * counts of members within evolutions. This function computes
 * the participation counts.
 * 
 * @param e List of communities constructing the evolution
 * @return A map that connects every vertex which appear in at least one
 *           community of e to the number of communities it is in.
 */
map < string, int > get_counts ( Evolution e ){
  map < string, int > result;
  map < string, int >::iterator it_r;

  for ( int i = 0; i < e.size(); i++ ){
    for ( int j = 0; j < e[i].size(); j++ ){
      if ( (it_r = result.find ( e[i][j] ) ) != result.end() ){
		it_r->second++;
      } else {
		result.insert ( pair < string, int > ( e[i][j], 1 ) );
      }
    }
  }

  return result;
}

/**
 * @fn void Print ( Community c )
 * 
 * Prints the members of the given community to standard out. Each 
 * member is space delimited.
 * 
 * @param C Community to print
 */
void Print ( Community c ){
  for ( int i = 0; i < c.size(); i++ ){
    cout << c[i] << " ";
  }
  cout << endl;
}

/**
 * @fn void Print ( Evolution e )
 * 
 * Prints an evolution to standard out. Each community gets a separate
 * line and the community members are space delimited.
 * 
 * @param e Evolution to print
 */
void Print ( Evolution e ){
  for ( int i = 0; i < e.size(); i++ ){
    Print ( e[i] );
  }
}

/**
 * @fn double M ( const map < string, int >& A, 
 *                          const map < string, int >& B )
 * 
 * Calculates the cosine similarity between participation vectors
 * of two evolutions. 
 * 
 * @param A,B Participation count of evolutions to compare
 * @return The cosine similarity between the vectors constructed with
 * 	         the participation counts.
 */
double M ( const map < string, int >& A, const map < string, int >& B ){
  map<string, int>::const_iterator it_a, it_b;
  double numerator = 0.0;
  double denominator = 0.0, bdenom = 0.0;

  for ( it_a = A.begin(); it_a != A.end(); it_a++ ) {
	//Sums squares of first evolution memberships
    denominator += (it_a->second) * (it_a->second);
    
    //Sum product of members between evolutions
    if ( ( it_b = B.find ( it_a->first ) ) != B.end() ){
      numerator += ( it_a->second * it_b->second );
    }
  }

  //Sums the squares for second evolution
  for ( it_b = B.begin(); it_b != B.end(); it_b++ ) {
    bdenom += ( it_b->second * it_b->second );
  }

  return ( numerator / ( sqrt ( denominator ) * sqrt ( bdenom ) ) );
}

typedef pair < int, int > id;

/**
 *@fn int main ( int argc, char** argv )
 * 
 * Main insertion point for evolution extension program
 * 
 * @param argc Number of command line arguments
 * @param argv Command line arguments
 * 
 * @return Not used
 */
int main ( int argc, char** argv ) {
	
  //Reads in command line arguments
  shared_ptr < Parameters > P ( new Parameters ( ) );
  P->Read( argc, argv );
  double threshold = P->get < double > ( "thr", 0.2 );
  vector < vector < Evolution > > starts, ends;

  ifstream fin;
  openFileHarsh ( &fin, P->get < string > ( "i", "detected_evolutions.dat" ) );
  
  vector < string > fields;
  int current_window = 0;
  
  map < id, id > finish_to_start;

  while ( fline_tr ( &fin, &fields, " " ) ){
	//Checks for a change in starting time step
    if ( fields[0].compare ( "Timestep" ) == 0 ){
      starts.push_back ( vector < Evolution > () );
      continue;
    } 
    //Checks for a delimiter for an evolution - this will only be hit 
    //  in weird cases (i.e. after a Timestep line) where the delimiter
    //  comes after an empty evolution.
    else if ( ( fields.size() == 1 ) && ( fields[0][0] == '-' ) ) {
      continue;
    }

	//It is guaranteed (if the input is formatted correctly) that the
	//lines in the file at this point are communities that belong to an
	//evolution. This loop constructs the full evolution - one community
	//per line
    Evolution next_evo;
    do {
      next_evo.push_back( fields );
    } while ( ( fline_tr ( &fin, &fields, " " ) ) && ( fields.size() > 1 ) && ( fields[0][0] != '-' ) );

	//Places the constructed evolution in the structure tracking 
	//    evolutions by start time window, the structure tracking
	//    evolutions by end time window, and the mapping in between the
	//    two.
    if ( next_evo.size() > 0 ){
      starts[starts.size() - 1].push_back ( next_evo );
      if ( ends.size() < starts.size() - 1 + next_evo.size() ){
		ends.resize ( starts.size() - 1 + next_evo.size() );
      }
      ends[starts.size() - 1 + next_evo.size() - 1].push_back ( next_evo );
      
      finish_to_start.insert ( pair < id, id > ( id ( starts.size() - 1 + next_evo.size() - 1, ends[starts.size() - 1 + next_evo.size() - 1].size() - 1 ), id ( starts.size() - 1, starts[starts.size() - 1].size() - 1 ) ) );
    }
  }
  fin.close();
 
  unsigned int extensions = 0;
  map < id, id > matches;
  //Iterates through all evolutions that end at a given time window.
  //Then compares the evolution to all evolutions that begin in the 
  //next time window. The best match is kept.
  
  //Easy future extension - change this so that instead of comparing 
  //  with only evolutions that begin in the following time window,
  //  it compares with evolutions that begin anywhere following the 
  //  current time window.
  for ( int i = ends.size() - 1; i >= 0; --i ){
    if ( ( i + 1 < starts.size() ) && ( starts[i+1].size() > 0 ) ){
      //Calculate M values and match
      vector < map < string, int > > start_counts;
      for ( int k = 0; k < starts[i+1].size(); k++ ){
		start_counts.push_back ( get_counts ( starts[i+1][k] ) );
      }
      
	  for ( int k = 0; k < ends[i].size(); k++ ){
		map < string, int > end_evo = get_counts ( ends[i][k] );
	
		double best_m = -1.0;
		int best_match = -1.0;
		for ( int t = 0; t < start_counts.size(); t++ ){
		  double this_m = M ( start_counts[t], end_evo );
		  if ( this_m>best_m ){
			best_m = this_m;
			best_match = t;
		  }
		}
		
		extensions++;
		if ( best_m > threshold ){
		  matches.insert ( pair < id, id > ( finish_to_start.find( id ( i, k ) )->second, id ( i+1, best_match ) ) );
		}
      }
    }
  }

  //Goes back through all the matches and prints out the resulting
  //   merged evolutions
  set < id > seen;
  for ( int i = 0; i < starts.size(); i++ ){
    cout << "Timestep " << i << endl;
    for ( int j = 0; j < starts[i].size(); j++ ){
      id current_id = id ( i, j );
      if ( seen.find ( current_id ) != seen.end() ) continue;
      
	  while ( true ){
		seen.insert ( current_id );
		Print ( starts[current_id.first][current_id.second] );
		
		if ( matches.find ( current_id ) != matches.end() ){
		  current_id = matches.find ( current_id )->second;
		} else {
		  break;
		}
      }
      cout << "----------" << endl;
    }
  }
}
